from ferret import Benchmark
from ferret import IntegratedGradientExplainer
from tqdm import tqdm
from abc import ABC, abstractmethod
import pandas as pd
import os
import torch


class Explainer(ABC):
    """ Abstract Explainer Class. """
    def __init__(self):
        return

    @abstractmethod
    def local_explanations(self, **kwargs):
        """ Abstract method to compute local explanations. """
        return

    def global_explanations(self, label_ids_to_explain, id2label, explainer_output_folder):
        """ Computes global explanations for a set of class labels.

        Args:
            label_ids_to_explain (list): List of class label ids to explain.
            id2label (dict): Dictionary mapping class label ids to class names.
            explainer_output_folder (str): Path to the folder containing the local explanations.

        Returns:
            dict: Dictionary containing the global explanations for each class label.
        """

        output_dict = {}

        for label_id in label_ids_to_explain:
            label_name = id2label[label_id]

            # Check if the directory contains at least one CSV file
            csv_files = [file for file in os.listdir(os.path.join(explainer_output_folder, "local_explanations", label_name)) if file.endswith(".csv")]
            if not csv_files:
                print(f"\nGlobal Explanations for label: Local explanations not found for label {label_name}. Skipping...")
                output_dict[label_name] = []
            else:
                df_current = self._load_local_explanations(os.path.join(explainer_output_folder, "local_explanations", label_name))
                df_current_global = self._compute_global_scores(df_current)
                df_current_global.to_csv(os.path.join(explainer_output_folder, "global_explanations", f"global_scores_{label_name}.csv"))
                print(f"\nGlobal Explanations for label: {label_name}")
                print(df_current_global)
                print("\n")

                output_dict[label_name] = df_current_global["tokens"].tolist()

        return output_dict

    @staticmethod
    def _load_local_explanations(dir_path) -> pd.DataFrame:
        """ Loads local explanations from a directory.

        Args:
            dir_path :obj:`str`: Path to the directory containing the local explanations.

        Returns:
            :obj:`pandas.core.frame.DataFrame`: Dataframe containing the local explanations.
        """

        df_list = []

        for path in os.listdir(dir_path):
            current_path = os.path.join(dir_path, path)
            if current_path.endswith(".csv"):
                df_tmp = pd.read_csv(current_path)
                df_list.append(df_tmp)

        if not df_list:
            raise ValueError(f"No CSV files found in directory: {dir_path}")

        if len(df_list) == 1:
            df = df_list[0]
        else:
            df = pd.concat(df_list, ignore_index=True)

        df["tokens"] = df["tokens"].astype(str)
        df["tokens"] = df["tokens"].apply(lambda x: x.replace(" ", ""))
        df["tokens"] = df["tokens"].apply(lambda x: x.lower())

        return df

    @staticmethod
    def _compute_global_scores(df, minimum_frequency=2, subtoken_separator="##") -> pd.DataFrame:
        """ Computes overall importance scores (global explanations) for each token by aggregating their importance withing individual predictions (local explanations).

        Args:
            df (pd.DataFrame): Dataframe containing local explanations.
            minimum_frequency (int, optional): Minimum frequency of a token to be considered in the global explanations. Defaults to 2.
            subtoken_separator (str, optional): Subtoken separator used by the tokenizer. Defaults to "##".

        Returns:
            pd.DataFrame: Dataframe containing the global explanations.
        """

        df_grp = df.groupby(["tokens"]).sum().reset_index()

        df_freq = df.groupby(["tokens"]).size().reset_index(name='freq')

        neg_freq_dict = {}

        for token, freq in zip(df_freq.tokens.tolist(), df_freq.freq.tolist()):
            neg_freq_dict[token] = freq

        df_global = df_grp.copy()

        df_global = df_global.join(df_freq.set_index('tokens'), on='tokens')

        df_global["score_norm"] = df_global["explanation_scores"] / df_global["freq"]
        df_global["score_norm_weighted"] = df_global["explanation_scores_weighted"] / df_global["freq"]

        df_global = df_global.drop(columns=["Unnamed: 0"])

        df_global = df_global.sort_values(by='score_norm', ascending=False)

        df_global = df_global.loc[df_global.freq >= minimum_frequency]

        df_global = df_global.loc[~df_global.tokens.str.startswith(subtoken_separator)]

        df_global = df_global.loc[df_global.tokens.str.len() > 2]

        df_global = df_global.loc[~df_global.tokens.str.isdigit()]

        df_global = df_global.loc[df_global.score_norm > 0]

        return df_global


class IntegratedGradientsExplainer(Explainer):
    """ Integrated Gradients implementation of the Explainer abstract class. """

    def __init__(self, model, tokenizer, device="cpu"):
        super().__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.model.to(device)

        # Create IG explainer Ferret library
        self.igx = IntegratedGradientExplainer(model, tokenizer)
        self.bench = Benchmark(model, tokenizer, explainers=[self.igx])
        return

    @staticmethod
    def batch_iterator(list1, list2, batch_size):
        """ Iterates two list by batch_size.

        Args:
            list1 (list): First list.
            list2 (list): Second list.
            batch_size (int): Batch size.
        Yields:
            tuple: Tuple containing the current batch of list1, list2, start_index and end_index.
        """

        # Ensure both input lists have the same length
        if len(list1) != len(list2):
            raise ValueError("Input lists must have the same length")

        # Iterate over the lists by batch_size
        for i in range(0, len(list1), batch_size):
            start_index = i
            end_index = min(i + batch_size, len(list1))
            # Return the current batch
            yield list1[start_index:end_index], list2[start_index:end_index], start_index, end_index

    def local_explanations(self, df_predictions, local_explanations_folder, label_ids_to_explain, id2label, batch_size=128):
        """ Computes and saves local explanations for a set of predictions in the output folder.

        Args:
            df_predictions (pd.DataFrame): Dataframe containing the predictions to explain.
            local_explanations_folder (str): Path to the folder where the local explanations will be saved.
            label_ids_to_explain (list): List of label ids to explain.
            id2label (dict): Dictionary mapping label ids to label names.
            batch_size (int, optional): Batch size. Defaults to 128.

        """

        log_errors = []
        log_filepath = os.path.join(local_explanations_folder, "log_file.txt")

        start_index = 0

        label_ids_to_explain
        # Iterate over labels to explain
        for label_id in label_ids_to_explain:
            label_name = id2label[label_id]

            # Create folder for current label
            current_path = os.path.join(local_explanations_folder, label_name)
            if not os.path.exists(current_path):
                os.mkdir(current_path)

            # Select texts predicted with the current label to explain
            df_predictions_current_label = df_predictions.loc[df_predictions.pred_label_name == label_name]

            print(f"Explainer: Explaining label: {label_name} - Number of texts: {len(df_predictions_current_label)}")

            # Iterate over batches of texts
            for batch_texts, batch_predicted_prob_explained_class, start_index, end_index in self.batch_iterator(df_predictions_current_label["text"].tolist(), df_predictions_current_label["pred_score"].tolist(), batch_size):
                try:
                    print(f"\t Explaining Batch: {start_index} -> {end_index}")

                    # Compute local explanation on current batch
                    batch_tokens, batch_scores, batch_scores_weighted = self._compute_batch_local_explanations(label_id, batch_texts, batch_predicted_prob_explained_class)

                    # Save local explanations of current batch on a csv file
                    df = pd.DataFrame({"tokens": batch_tokens, "explanation_scores": batch_scores, "explanation_scores_weighted": batch_scores_weighted})
                    df.to_csv(os.path.join(current_path, f"scores_{start_index}_{end_index}.csv"))

                except Exception as e:
                    print(f"An exception occurred in {start_index} -> {end_index}")
                    print(e)
                    log_errors.append(f"An exception occurred in {start_index} -> {end_index} in label {label_name}")
                    with open(log_filepath, "a") as file_object:
                        file_object.write(f"\nAn exception occurred in {start_index} -> {end_index} in label {label_name}")

        return

    def _compute_batch_local_explanations(self, label_id, batch_texts, batch_pred_scores) -> tuple[list[str], list[float], list[float]]:
        """ Computes local explanations for a batch of texts.

        Args:
            label_id (:obj:`int`): Label id to explain.
            batch_texts (:obj:`list[str]`): List of texts to explain.
            batch_pred_scores (:obj:`list[float]`): List of predicted probabilities for the target label.

        Returns:
            :obj:`tuple(list(str), list(float), list(float))`: Tuple containing the tokens, feature importance scores, and weighted feature importance scores.
        """

        batch_tokens = []
        batch_scores = []
        batch_scores_weighted = []

        for text, prob in tqdm(zip(batch_texts, batch_pred_scores)):
            text = str(text)

            #tokenized_text = self.tokenizer.tokenize(text)
            #if len(tokenized) > 320:
            #    text = self.tokenizer.convert_tokens_to_string(tokenized[:320])

            explanations = self.bench.explain(text, target=label_id, show_progress=False, normalize_scores=True)

            # Remove [CLS] and [SEP] tokens
            tokens = explanations[0].tokens[1:-1]
            scores_raw = explanations[0].scores[1:-1].tolist()

            # Produce importance score weighted by predicted probability
            scores_weighted = [s * prob for s in scores_raw]

            # Store tokens, feature importance, and weighted feature importance
            batch_tokens.extend(tokens)
            batch_scores.extend(scores_raw)
            batch_scores_weighted.extend(scores_weighted)
        return batch_tokens, batch_scores, batch_scores_weighted


class ShapExplainer(Explainer):
    """ Shap implementation of the Explainer abstract class. """

    def local_explanations(self):
        return

    def global_explanations(self):
        return




from abc import ABC, abstractmethod
import pandas as pd
import nltk
from nltk.corpus import wordnet
import gensim.downloader as api
import random


class Moderator(ABC):
    """ Abstract Moderator Class. """
    def __init__(self):
        return

    @abstractmethod
    def words_removal_mitigation_strategy(self, **kwargs):
        """ Abstract method to perform the word removal mitigation strategy. """

        return

    @abstractmethod
    def sentences_removal_mitigation_strategy(self, **kwargs):
        """ Abstract method to perform the sentence removal mitigation strategy. """

        return

    @abstractmethod
    def word_replacement_with_synonyms_mitigation_strategy(self, **kwargs):
        """ Abstract method to perform the word replacement with synonyms mitigation strategy. """

        return

    @abstractmethod
    def word_replacement_with_hypernym_mitigation_strategy(self, **kwargs):
        """ Abstract method to perform the word replacement with hypernym mitigation strategy. """

        return


class PandasDataFrameModerator(Moderator):
    """ Moderator Implementation for Pandas DataFrames. """
    def __init__(self):
        super().__init__()
        return

    def words_removal_mitigation_strategy(self, df_train, tokenizer, protected_attributes_per_label_dict, text_column_name, label_column_name, id2label, mitigate_each_label_separately=False, batch_size=128) -> pd.DataFrame:
        """ Performs the word removal mitigation strategy.

        Args:
            df_train (:obj:`pandas.DataFrame`): Training dataset.
            tokenizer (:obj:`transformers.AutoTokenizer`): Tokenizer.
            protected_attributes_per_label_dict (:obj:dict): Dictionary of protected attributes per class label.
            text_column_name (:obj:`str`): Name of the column containing the text.
            label_column_name (:obj:`str`): Name of the column containing the class label.
            id2label (:obj:`dict`): Dictionary mapping class label ids to class label names
            id2label (:obj:`dict`): Dictionary mapping class label ids to class label names.
            mitigate_each_label_separately (:obj:`bool`, `optional`): Whether to mitigate each class label separately.
            batch_size (:obj:`int`, `optional`): Batch size for processing. Default is 128.

        Returns:
            :obj:`pandas.DataFrame`: DataFrame with the mitigated texts.
        """

        if mitigate_each_label_separately == False:
            # Remove words related to protected attributes from all labels
            distinct_protected_attributes = list(set(word for words_list in protected_attributes_per_label_dict.values() for word in words_list))

            # Apply the mitigation strategy in batches to all class labels
            df_train['mitigated_text'] = pd.concat([pd.Series(self._batch_words_removal(df_train[text_column_name][i:i + batch_size], tokenizer, distinct_protected_attributes)) for i in range(0, df_train.shape[0], batch_size)], ignore_index=True)
        else:
            # Apply the mitigation strategy in batches to each class label separately
            for label_id in protected_attributes_per_label_dict.keys():
                # Remove words related to protected attributes from current label
                distinct_protected_attributes = list(set(word for words_list in protected_attributes_per_label_dict[label_id] for word in words_list))

                label_name = id2label[label_id]

                # Apply the mitigation strategy in batches to current class label
                df_train.loc[df_train[label_column_name] == label_name, 'mitigated_text'] = pd.concat([pd.Series(self._batch_words_removal(df_train.loc[df_train[label_column_name] == label_name, text_column_name][i:i + batch_size], tokenizer, distinct_protected_attributes)) for i in range(0, df_train.loc[df_train[label_column_name] == label_name].shape[0], batch_size)], ignore_index=True)
        return df_train

    @staticmethod
    def _batch_words_removal(texts, tokenizer, protected_attributes):
        """ Removes words from texts in batch.

        Args:
            texts (:obj:`list(str)`): List of texts.
            tokenizer (:obj:`transformers.AutoTokenizer`): Tokenizer.
            protected_attributes (:obj:`list` of :obj:`str`): List of protected attributes to remove.

        Returns:
            :obj:`list(str)`: List of mitigated texts.
        """

        # Tokenize the texts in batch
        encoded_inputs = tokenizer(texts.tolist(), add_special_tokens=False, return_tensors='pt', padding=True)

        # Process each text
        clean_texts = []
        for i in range(encoded_inputs.input_ids.size(0)):
            # Ensure input_ids[i] is 1-dimensional
            input_ids = encoded_inputs.input_ids[i]
            if input_ids.dim() == 0:
                input_ids = input_ids.unsqueeze(0)

            # Convert token IDs to tokens
            tokens = [tokenizer.convert_ids_to_tokens(id.item()) for id in input_ids]

            # Remove unwanted tokens and reconstruct the text
            tokens = [t for t in tokens if t not in protected_attributes]
            clean_texts.append(tokenizer.convert_tokens_to_string(tokens))

        return clean_texts

    @staticmethod
    def _batch_sentences_removal(texts, tokenizer, protected_attributes):
        """ Removes sentences containing protected attributes from texts in batch.

        Args:
            texts (:obj:`list` of :obj:`str`): List of texts.
            tokenizer (:obj:`transformers.AutoTokenizer`): Tokenizer.
            protected_attributes (:obj:`list` of :obj:`str`): List of protected attributes to remove.

        Returns:
            :obj:`list(str)`: List of mitigated texts obtained by removing sentences containing protected attributes.
        """

        # Tokenize the texts in batch
        encoded_inputs = tokenizer(texts.tolist(), add_special_tokens=False, return_tensors='pt', padding=True)

        # Process each text
        clean_texts = []
        for i in range(encoded_inputs.input_ids.size(0)):
            # Ensure input_ids[i] is 1-dimensional
            input_ids = encoded_inputs.input_ids[i]
            if input_ids.dim() == 0:
                input_ids = input_ids.unsqueeze(0)

            # Convert token IDs to tokens
            tokens = [tokenizer.convert_ids_to_tokens(id.item()) for id in input_ids]

            contains_protected_attribute = False
            for word in protected_attributes:
                if word in tokens:
                    contains_protected_attribute = True
                    break

            if contains_protected_attribute:
                clean_texts.append(None)
            else:
                clean_texts.append(texts.tolist()[i])

        return clean_texts

    def sentences_removal_mitigation_strategy(self, df_train, tokenizer, protected_attributes_per_label_dict, text_column_name, label_column_name, id2label, mitigate_each_label_separately=False, batch_size=128) -> pd.DataFrame:

        """ Performs the sentence removal mitigation strategy.

        Args:
            df_train (:obj:`pandas.DataFrame`): Training dataset.
            tokenizer (:obj:`transformers.AutoTokenizer`): Tokenizer.
            protected_attributes_per_label_dict (:obj:`dict`): Dictionary of protected attributes per class label.
            text_column_name (:obj:`str`): Name of the column containing the text.
            label_column_name (:obj:`str`): Name of the column containing the class label.
            id2label (:obj:`dict`): Dictionary mapping class indices to class labels.
            mitigate_each_label_separately (:obj:`bool`, optional): Whether to mitigate each class label separately. If `True` Protected attributes identified as important for a class label are mitigated for that class label only. If `False` protected attributes identified for a particular class label are mitigated for all class labels. Defaults to `False`.
            batch_size (:obj:`int`, optional): Batch size. Defaults to 128.

        Returns:
            :obj:`pandas.DataFrame`: Dataframe with mitigated texts obtained by protected attributes sentence removal.
        """

        if mitigate_each_label_separately == False:
            # Remove sentences related to protected attributes from all labels
            distinct_protected_attributes = list(set(word for words_list in protected_attributes_per_label_dict.values() for word in words_list))

            # Apply the mitigation strategy in batches to all class labels
            df_train['mitigated_text'] = pd.concat([pd.Series(self._batch_sentences_removal(df_train[text_column_name][i:i + batch_size], tokenizer, distinct_protected_attributes)) for i in range(0, df_train.shape[0], batch_size)], ignore_index=True)
        else:
            # Apply the mitigation strategy in batches to each class label separately
            for label_id in protected_attributes_per_label_dict.keys():
                # Remove words related to protected attributes from current label
                distinct_protected_attributes = list(set(word for words_list in protected_attributes_per_label_dict[label_id] for word in words_list))

                label_name = id2label[label_id]

                # Apply the mitigation strategy in batches to current class label
                df_train.loc[df_train[label_column_name] == label_name, 'mitigated_text'] = pd.concat([pd.Series(self._batch_sentences_removal(df_train.loc[df_train[label_column_name] == label_name, text_column_name][i:i + batch_size], tokenizer, distinct_protected_attributes)) for i in range(0, df_train.loc[df_train[label_column_name] == label_name].shape[0], batch_size)], ignore_index=True)
        return df_train

    def word_replacement_with_synonyms_mitigation_strategy(self, df_train, tokenizer,
                                                           protected_attributes_per_label_dict, text_column_name,
                                                           label_column_name, id2label, n_synonyms=5,
                                                           keep_original_sentence=True,
                                                           mitigate_each_label_separately=False, batch_size=128) -> pd.DataFrame:
        """ Performs the word replacement with synonyms mitigation strategy. `n_synonyms` synonyms are generated for each protected attributes in each text by replacing each protected attributes with one of the most similar `n_synonyms` words.

        Args:
            df_train (:obj:`pandas.DataFrame`): Training dataset.
            tokenizer (:obj:`transformers.AutoTokenizer`): Tokenizer.
            protected_attributes_per_label_dict (:obj:`dict`): Dictionary of protected attributes per class label.
            text_column_name (:obj:`str`): Name of the column containing the text.
            label_column_name (:obj:`str`): Name of the column containing the class label.
            id2label (:obj:`dict`): Dictionary mapping class indices to class labels.
            n_synonyms (:obj:`int`, optional): Number of synonym-based sentences to generate per text. Defaults to 5.
            keep_original_sentence (:obj:`bool`, optional): Whether to keep the original sentence containing the identified protected attribute. Defaults to True.
            mitigate_each_label_separately (:obj:`bool`, optional): Whether to mitigate each class label separately. If `True` Protected attributes identified as important for a class label are mitigated for that class label only. If `False` protected attributes identified for a particular class label are mitigated for all class labels. Defaults to `False`.
            batch_size (:obj:`int`, optional): Batch size. Defaults to 128.

        Returns:
            :obj:`pandas.DataFrame`: Dataframe with mitigated texts obtained by replacing protected attributes with synonyms.
        """

        glove_word_embedding = self.load_GloVe_embedding_model()

        if mitigate_each_label_separately:
            for label_id, protected_words in protected_attributes_per_label_dict.items():
                label_name = id2label[label_id]
                label_df = df_train[df_train[label_column_name] == label_name]
                df_train.loc[df_train[label_column_name] == label_name, 'synonym_texts'] = pd.concat(
                    [pd.Series(self._generate_synonym_sentences(
                        label_df[text_column_name][i:i + batch_size],
                        protected_words,
                        glove_word_embedding,
                        n_synonyms,
                        keep_original_sentence
                    )) for i in range(0, label_df.shape[0], batch_size)], ignore_index=True
                )
        else:
            all_protected_words = list(
                set(word for words in protected_attributes_per_label_dict.values() for word in words))
            df_train['synonym_texts'] = pd.concat(
                [pd.Series(self._generate_synonym_sentences(
                    df_train[text_column_name][i:i + batch_size],
                    all_protected_words,
                    glove_word_embedding,
                    n_synonyms,
                    keep_original_sentence
                )) for i in range(0, df_train.shape[0], batch_size)], ignore_index=True
            )
        return df_train

    def _generate_synonym_sentences(self, texts, protected_words, glove_word_embedding, n_synonyms,
                                    keep_original_sentence) -> list[str]:
        """ Generate sentences by replacing words with synonyms.

        Args:
            texts (:obj:`list(str)`): List of original sentences.
            protected_words (:obj:`list(str)`): List of protected words.
            glove_word_embedding (:obj:`gensim.models.KeyedVectors`): GloVe embeddings model.
            n_synonyms (:obj:`int`): Number of synonym sentences to generate.
            keep_original_sentence (:obj:`bool`): Flag indicating weather to keep the original sentence in the output.

        Returns:
            :obj:`list(str)`: List of texts with synonym-replaced sentences for each original sentence.
        """

        synonym_sentences = []

        # Generate synonyms for the entire batch of protected words once
        synonyms_dict = self._get_synonyms(protected_words, glove_word_embedding, k=5)
        print("Synonym Dictionary:", synonyms_dict)  # Debugging line

        for text in texts:
            tokenized_text = text.split()  # Tokenize sentence into words

            # Start with the original sentence if desired
            sentence_variants = [text] if keep_original_sentence else []

            # Generate n_synonyms variations
            for _ in range(n_synonyms):
                new_sentence = []
                for word in tokenized_text:
                    # Replace word with a random synonym if it's a protected word
                    if word in protected_words and word in synonyms_dict:
                        synonym_options = synonyms_dict[word]
                        new_word = random.choice(synonym_options)

                        # Ensure new synonym is different from original to avoid duplicates
                        if new_word == word and len(synonym_options) > 1:
                            synonym_options = [syn for syn in synonym_options if syn != word]
                            new_word = random.choice(synonym_options)

                        new_sentence.append(new_word)

                        print(f"replacing word {word} with sysnonym {new_word} in:")
                        print(new_sentence)
                        print("\n\n")
                    else:
                        new_sentence.append(word)

                # Join words to form the new sentence and add it to sentence_variants
                sentence_variants.append(" ".join(new_sentence))

            synonym_sentences.append(sentence_variants)

        return synonym_sentences


    def word_replacement_with_hypernym_mitigation_strategy(self, **kwargs):
        # TODO
        # Download the WordNet corpus
        nltk.download('wordnet')

        self._get_hypernyms()
        return

    @staticmethod
    def _get_synonyms(word_list, glove_word_embedding, k=5) -> dict[str, list[str]]:
        """ Returns synonyms for the given word list using the GloVe embeddings.

        Args:
            :obj:`list(str)`: List of words to find synonyms for.
            :obj:`gensim.models.KeyedVectors`: Pre-trained word embeddings.
            :obj:(`int`, `optional`): Number of synonyms to return for each word. Defaults to 5.

        Returns:
            :obj:`dict`: Dictionary of words and their corresponding synonyms.
        """

        synonyms_dict = {}
        for word in word_list:
            if word in glove_word_embedding:
                k_similar_words = [similar_word[0] for similar_word in glove_word_embedding.most_similar(word, topn=k)]
                synonyms_dict[word] = k_similar_words
            else:
                synonyms_dict[word] = [word]  # Fallback to original word if no synonyms are found
        return synonyms_dict

    @staticmethod
    def load_GloVe_embedding_model(model_name="glove-wiki-gigaword-300"):
        """ Load the GloVe embeddings model.

        Args:
            model_name (str): Name of the GloVe model to load.

        Returns:
            gensim.models.KeyedVectors: Loaded GloVe embeddings.
        """

        return api.load(model_name)

    @staticmethod
    def _get_hypernyms(word_list):
        """" Returns the hypernyms of the words in the given list. """

        hypernyms_dict = {}
        for word in word_list:
            print(word)
            my_syns = [wordnet.synsets(word)[0]]
            if len(my_syns) > 0:
                ss_list = []
                for ss in my_syns:
                    ss_list.extend(ss.hypernyms())
                if len(ss_list) > 0:
                    hypernyms_dict[word] = ss_list[0].name()

        for w, h in hypernyms_dict.items():
            tmp_h = h.split(".", 1)[0]
            tmp_h = tmp_h.replace("_", " ")
            hypernyms_dict[w] = tmp_h
        return hypernyms_dict


class HuggingFaceDatasetModerator(Moderator):
    # TODO - Implement the HUGGINGFACE Dataset Moderator
    def __init__(self):
        super().__init__()
        return

    def words_removal_mitigation_strategy(self, **kwargs):
        return

    def sentences_removal_mitigation_strategy(self, **kwargs):
        return

    def word_replacement_with_synonyms_mitigation_strategy(self, **kwargs):
        return

    def word_replacement_with_hypernym_mitigation_strategy(self, **kwargs):
        return



from abc import ABC, abstractmethod
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM, AutoModelForCausalLM
from tqdm import tqdm
import pandas as pd
import time
import datetime
import openai
import torch

class Identifier(ABC):
    """ Abstract Moderator Class. """
    def __init__(self):
        return

    @abstractmethod
    def annotate_protected_attributes(self, **kwargs):
        """ Abstract method that annotates the protected attributes. """
        return


class ChatGPTIdentifier(Identifier):
    def __init__(self, openai_api_key=""):
        super().__init__()
        openai.api_key = openai_api_key
        return

    def annotate_protected_attributes(self, tokens, temperature=0.3):
        """  ."""
        result = []
        raw_responses = []
        exception_logs = []

        for tk in tqdm(tokens):
            chatgpt_response = ""
            try:
                print(tk)
                chatgpt_response = self._chatgpt_annotate(tk)
                print(chatgpt_response)
                raw_responses.append(f"{tk}: {chatgpt_response}\n")
                chatgpt_main_category, chatgpt_category_main_score, chatgpt_explanation = chatgpt_response.split('|')

                result.append({'word': tk,
                               'chatgpt_protected_category': chatgpt_main_category,  # Protected category annotated by chatgpt
                               'chatgpt_protected_category_score': chatgpt_category_main_score,  # Reliability score of the protected category
                               'chatgpt_explanation': chatgpt_explanation  # Explanation provided for tha annotated category
                               })
            except Exception as ex:
                try:
                    chatgpt_response = chatgpt_response.split('\n')[2]
                    chatgpt_main_category, chatgpt_category_main_score, chatgpt_explanation = chatgpt_response.split('|')
                    result.append({'word': tk,
                                   'chatgpt_protected_category': chatgpt_main_category,  # Protected category annotated by chatgpt
                                   'chatgpt_protected_category_score': chatgpt_category_main_score,  # Reliability score of the protected category
                                   'chatgpt_explanation': chatgpt_explanation  # Explanation provided for tha annotated category
                                   })
                except Exception as ex:
                    print(f"{ex} in {tk}!")
                    print({chatgpt_response})
                    print("\n")
                    exception_logs.append(f"Exception in: {tk}: {chatgpt_response}\n")

        ts = time.time()
        timestamp = datetime.datetime.fromtimestamp(ts).strftime('%Y%m%d_%H%M%S')
        filename = f"ChatGPT-annotation-{timestamp}"

        df_annotated = pd.DataFrame(result)

        df_cleaned = self._clean_chatgpt_responses(df_annotated)

        result, acc_list = self._aggregate_protected_categories_votes(df_cleaned)

        print(result)

        protected_attributes = result.loc[result.count_pa > result.count_non_pa].word.tolist()



        print(f"Identifier - Identified Protected Attributes: {protected_attributes}")

        # TODO - Add logic to save annotation and create a dynamic dictionary of protected attributes
        #df_annotated.to_csv(f"chatgpt_annotation/{filename}.csv", index=None)

        #with open(f'chatgpt_annotation/raw_responses/{filename}.txt', 'w') as responses_file:
        #    responses_file.writelines(raw_responses)

        #with open(f'chatgpt_annotation/log/{filename}.txt', 'w') as responses_file:
        #    responses_file.writelines(exception_logs)

        return result, protected_attributes

    def _clean_chatgpt_responses(self, df_raw,  protected_category_column_name="chatgpt_protected_category"):
        """ Cleans the chatgpt response.
        Args:
            df (pd.DataFrame): The dataframe containing the chatgpt response.
        """
        df = df_raw.copy()
        df["word"] = df["word"].str.lower()
        df["word"] = df["word"].str.strip()
        df[protected_category_column_name] = df[protected_category_column_name].str.lower()
        df[protected_category_column_name] = df[protected_category_column_name].str.strip()

        # Replace the "None" category with "none-category"
        df[protected_category_column_name].replace('none', 'none-category', inplace=True)
        df[protected_category_column_name] = df[protected_category_column_name].fillna("none-category")

        # Correct errors in the response provided by chatgpt for some categories
        df[protected_category_column_name].replace('"pregnancy and maternity', 'pregnancy and maternity', inplace=True)
        df[protected_category_column_name].replace('"disability', 'disability', inplace=True)
        df[protected_category_column_name].replace('"race', 'race', inplace=True)
        return df

    @staticmethod
    def _aggregate_protected_categories_votes(df, protected_category_column_name="chatgpt_protected_category"):
        # Compute the counts in a single grouped operation
        grouped = df.groupby('word')[protected_category_column_name].value_counts().unstack(fill_value=0)

        # Add a total count column
        grouped['count_tot'] = grouped.sum(axis=1)

        # Check if 'none-category' column exists and rename it to 'count_non_pa'
        if 'none-category' in grouped.columns:
            grouped = grouped.rename(columns={'none-category': 'count_non_pa'})
            # Calculate 'count_pa' (which is count_tot - count_non_pa)
            grouped['count_pa'] = grouped['count_tot'] - grouped['count_non_pa']
        else:
            # If 'none-category' column doesn't exist, set 'count_non_pa' to 0 and 'count_pa' to 'count_tot'
            grouped['count_non_pa'] = 0
            grouped['count_pa'] = grouped['count_tot']

        # Reset the index to turn 'word' back into a column
        grouped = grouped.reset_index()

        # Select only the required columns
        result = grouped[['word', 'count_pa', 'count_non_pa', 'count_tot']]

        # Convert to a list of dictionaries, if needed
        acc_list = result.to_dict('records')
        return result, acc_list

    @staticmethod
    def _chatgpt_annotate(tk, temperature=0.3, chatgpt_model="gpt-3.5-turbo"):
        """ Performs annotation using ChatGPT of the input token.
        Args:
            tk (str): The token to be annotated.
            temperature (float): The temperature of the ChatGPT model.
            chatgpt_model (str): The ChatGPT model to be used.
        """
        completion = openai.chat.completions.create(
            model=chatgpt_model,
            messages=[
                {'role': 'user', 'content': '''Consider these 9 protected categories defined by the Equality Act law to avoid discrimination of automatic decision-making algorithms:
                "Age": A person belonging to a particular age (for example, 32 year olds) or range of ages (for example 18 to 30 year olds).
                "Disability": A person has a disability if she or he has a physical or mental impairment which has a substantial and long-term adverse effect on that person's ability to carry out normal day-to-day activities.
                "Gender reassignment": The process of transitioning from one sex to another.
                "Marriage and civil partnership": Marriage is a union between a man and a woman or between a same-sex couple. Same-sex couples can also have their relationships legally recognised as 'civil partnerships'. Civil partners must not be treated less favourably than married couples.
                "Pregnancy and maternity": Pregnancy is the condition of being pregnant or expecting a baby. Maternity refers to the period after the birth, and is linked to maternity leave in the employment context. In the non-work context, protection against maternity discrimination is for 26 weeks after giving birth, and this includes treating a woman unfavourably because she is breastfeeding.
                "Race": Refers to the protected characteristic of race. It refers to a group of people defined by their race, colour, and nationality (including citizenship) ethnic or national origins.
                "Religion and belief": Religion refers to any religion, including a lack of religion. Belief refers to any religious or philosophical belief and includes a lack of belief. Generally, a belief should affect your life choices or the way you live for it to be included in the definition.
                "Sex": A word that explicitly refers to the gender of a person: e.g., man or woman, she or he, mr or mrs, male of female, madame etc.
                "Sexual orientation": Whether a person's sexual attraction is towards their own sex, the opposite sex or to both sexes.
                '''},
                {'role': 'user', 'content': """You can learn more about the discriminations along each protected characteristic on the following URLs:
                "Age" : https://www.equalityhumanrights.com/en/advice-and-guidance/age-discrimination
                "Disability":https://www.equalityhumanrights.com/en/disability-advice-and-guidance
                "Gender reassignment": https://www.equalityhumanrights.com/en/advice-and-guidance/gender-reassignment-discrimination
                "Marriage and civil partnership": https://www.equalityhumanrights.com/en/advice-and-guidance/marriage-and-civil-partnership-discrimination
                "Pregnancy and maternity": https://www.equalityhumanrights.com/en/node/5916
                "Race": https://www.equalityhumanrights.com/en/advice-and-guidance/race-discrimination
                "Religion and belief": https://www.equalityhumanrights.com/en/religion-or-belief-work
                "Sex": https://www.equalityhumanrights.com/en/advice-and-guidance/sex-discrimination
                "Sexual orientation": https://www.equalityhumanrights.com/en/advice-and-guidance/sexual-orientation-discrimination   """},

                {'role': 'user', 'content': f"""Given the previously defined protected categories "Age", "Disability", "Gender reassignment", "Marriage and civil partnership", "Pregnancy and maternity", "Race", "Religion and belief", "Sex", "Sexual orientation". 
                How would you classify the word "{tk}" and which [0,100] reliability scores (only one) would you give to your assessment? You must assign one category. 
                If a word does not fit any categories, you must assign the category "None" with the reliability score and the relative explanation. 
                Provide the answer in the format: "Protected Category|Reliability Score from 0 to 100 for the protected category|Explanation of why the word belong to the protected category.". 
                In case of a word does not fall into any category, provide the answer in the format: "None|Reliability Score from 0 to 100 for the None category|Explanation of why the word does not fall under any of the defined protected categories. 
                Each answer MUST have exactly two | symbols and only one line; otherwise, I cannot process your response. 
                Please annotate a word to a protected category only if it is strictly related. """}
            ],
            temperature=temperature,
            max_tokens=150
        )
        print(completion)
        return completion.choices[0].message.content


class LLamaIdentifier(Identifier):
    def __init__(self, model_name="meta-llama/Meta-Llama-3-8B-Instruct", hf_token="",  device='cuda' if torch.cuda.is_available() else 'cpu'):
        super().__init__()
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, use_auth_token=hf_token)
        self.model = AutoModelForCausalLM.from_pretrained(model_name, use_auth_token=hf_token).to(device)
        self.device = device

    def annotate_protected_attributes(self, tokens, temperature=0.3):
        """ Annotates tokens with protected attributes using LLaMA-based model """
        results = []
        raw_responses = []
        exception_logs = []

        for tk in tqdm(tokens):
            try:
                llama_response = self._llama_annotate(tk, temperature)
                raw_responses.append(f"{tk}: {llama_response}\n")

                # Parse response
                protected_category, reliability_score, explanation = llama_response.split('|')
                results.append({
                    'word': tk,
                    'llama_protected_category': protected_category.strip(),
                    'llama_protected_category_score': reliability_score.strip(),
                    'llama_explanation': explanation.strip()
                })
            except Exception as ex:
                exception_logs.append(f"Exception for token '{tk}': {str(ex)}\n")

        # Save results to dataframe and log exceptions
        timestamp = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
        df_annotated = pd.DataFrame(results)

        # Clean and aggregate responses
        df_cleaned = self._clean_llama_responses(df_annotated)
        result, acc_list = self._aggregate_protected_categories_votes(df_cleaned)

        # Log the identified protected attributes
        protected_attributes = result[result.count_pa > result.count_non_pa].word.tolist()
        print(f"LLaMAIdentifier - Identified Protected Attributes: {protected_attributes}")

        return result, protected_attributes

    def _llama_annotate(self, tk, temperature=0.3, prompt_template=None):
        """ Generates annotation for a token using LLaMA """
        if prompt_template is None:
            prompt_template = (
                f"""Consider these 9 protected categories defined by the Equality Act law to avoid discrimination of automatic decision-making algorithms:
                "Age": A person belonging to a particular age (for example, 32 year olds) or range of ages (for example 18 to 30 year olds).
                "Disability": A person has a disability if she or he has a physical or mental impairment which has a substantial and long-term adverse effect on that person's ability to carry out normal day-to-day activities.
                "Gender reassignment": The process of transitioning from one sex to another.
                "Marriage and civil partnership": Marriage is a union between a man and a woman or between a same-sex couple. Same-sex couples can also have their relationships legally recognised as 'civil partnerships'. Civil partners must not be treated less favourably than married couples.
                "Pregnancy and maternity": Pregnancy is the condition of being pregnant or expecting a baby. Maternity refers to the period after the birth, and is linked to maternity leave in the employment context. In the non-work context, protection against maternity discrimination is for 26 weeks after giving birth, and this includes treating a woman unfavourably because she is breastfeeding.
                "Race": Refers to the protected characteristic of race. It refers to a group of people defined by their race, colour, and nationality (including citizenship) ethnic or national origins.
                "Religion and belief": Religion refers to any religion, including a lack of religion. Belief refers to any religious or philosophical belief and includes a lack of belief. Generally, a belief should affect your life choices or the way you live for it to be included in the definition.
                "Sex": A word that explicitly refers to the gender of a person: e.g., man or woman, she or he, mr or mrs, male of female, madame etc.
                "Sexual orientation": Whether a person's sexual attraction is towards their own sex, the opposite sex or to both sexes.
                .\n"""
                f"Classify the word '{tk}' into one of these nine categories or None if it does not belong to any, provide a Reliability Score 0-100, and an explanation for your annotation. "
                f"Provide in format: Protected Category|Reliability Score 0-100|Explanation.")

        # Encode prompt and generate response
        inputs = self.tokenizer(prompt_template, return_tensors="pt").to(self.device)
        outputs = self.model.generate(
            **inputs, max_length=150, temperature=temperature, do_sample=True
        )

        # Decode and return the generated response text
        response = self.tokenizer.decode(outputs[0], skip_special_tokens=True)
        return response

    def _clean_llama_responses(self, df_raw, protected_category_column_name="llama_protected_category"):
        """ Cleans and standardizes the responses for analysis. """
        df = df_raw.copy()
        df["word"] = df["word"].str.lower().str.strip()
        df[protected_category_column_name] = df[protected_category_column_name].str.lower().str.strip()

        # Replace "None" responses consistently
        df[protected_category_column_name].replace('none', 'none-category', inplace=True)
        df[protected_category_column_name].fillna("none-category", inplace=True)

        # Correct common errors if any
        df[protected_category_column_name].replace('"disability', 'disability', inplace=True)
        df[protected_category_column_name].replace('"race', 'race', inplace=True)

        return df

    @staticmethod
    def _aggregate_protected_categories_votes(df, protected_category_column_name="llama_protected_category"):
        # Aggregate counts for each word’s assigned categories
        grouped = df.groupby('word')[protected_category_column_name].value_counts().unstack(fill_value=0)
        grouped['count_tot'] = grouped.sum(axis=1)

        # Rename "none-category" to "count_non_pa" and compute "count_pa"
        if 'none-category' in grouped.columns:
            grouped.rename(columns={'none-category': 'count_non_pa'}, inplace=True)
            grouped['count_pa'] = grouped['count_tot'] - grouped['count_non_pa']
        else:
            grouped['count_non_pa'] = 0
            grouped['count_pa'] = grouped['count_tot']

        grouped.reset_index(inplace=True)
        result = grouped[['word', 'count_pa', 'count_non_pa', 'count_tot']]
        acc_list = result.to_dict('records')
        return result, acc_list




from nlpguard.explainer.explainer import IntegratedGradientsExplainer, ShapExplainer
from nlpguard.identifier.identifier import ChatGPTIdentifier, LLamaIdentifier
from nlpguard.moderator.moderator import PandasDataFrameModerator
from transformers.pipelines import pipeline
import os
import time
import datetime
from tqdm import tqdm
import pandas as pd
import torch


class NLPGuard:
    """ NLPGuard Framework class. It is used to run the mitigation framework.
    It is composed of three main components: Explainer, Identifier and Moderator.
    The Explainer is used to extract the most important words from a corpus of texts.
    The Identifier is used to identify the protected attributes from the most important words extracted by the explainer.
    The Moderator is used to mitigate the protected attributes identified by the identifier in a corpus of texts.

    Attributes:
        use_case_name (:obj:`str`): Name of the use case. It is used to create a new folder in the output folder.
        output_folder (:obj:`str`): Path to the output folder.
        explainer_output_folder (:obj:`str`): Path to the explainer output folder.
        identifier_output_folder (:obj:`str`): Path to the identifier output folder.
        moderator_output_folder (:obj:`str`): Path to the moderator output folder.
        id2label (:obj:`dict`): Dictionary mapping label ids to label names.
        label2id (:obj:`dict`): Dictionary mapping label names to label ids.

    """
    def __init__(self):
        self.use_case_name = None
        self.output_folder = None
        self.explainer_output_folder = None
        self.identifier_output_folder = None
        self.moderator_output_folder = None
        self.id2label = None
        self.label2id = None
        return

    def initialize_mitigation_framework(self, id2label, use_case_name=None, output_folder=None) -> None:
        """ Initializes the output folders.

            Args:
                id2label (:obj:`dict`): Dictionary mapping label ids to label names.
                use_case_name (:obj:`str`): Name of the use case. It is used to create a new folder in the output folder.
                output_folder (:obj:`str`): Path to the output folder.

            Returns:
                None
        """
        self.id2label = id2label
        self.label2id = {v: k for k, v in id2label.items()}
        self.output_folder, self.use_case_name, self.explainer_output_folder, self.identifier_output_folder, self.moderator_output_folder = self._init_output_folders(use_case_name, output_folder)
        return

    @staticmethod
    def _init_output_folders(use_case_name, output_folder) -> tuple:
        """ Initializes the output folders.

        Args:
            use_case_name (:obj:`str`): Name of the use case. It is used to create a new folder in the output folder.
            output_folder (:obj:`str`): Path to the output folder.

        Returns:
            :obj:`tuple(str)`: Tuple containing the output folder, the use case folder, the explainer output folder, the identifier output folder and the moderator output folder.
        """
        # Init output folder. If output folder is not specified, the dafault is "outputs".
        if output_folder is None:
            output_folder = ("outputs")

        # Create the output folder if it does not exist.
        if not os.path.exists(output_folder):
            print("NLPGuard: Warning, output folder does not exist.")
            os.mkdir(output_folder)
            print("NLPGuard: Output folder created.")

        # Init use case folder. If use_case_name is not specified, the dafault is the current timestamp.
        if use_case_name is None:
            ts = time.time()
            timestamp = datetime.datetime.fromtimestamp(ts).strftime('%Y%m%d_%H%M%S')
            use_case_name = timestamp
        else:
            if os.path.exists(os.path.join(output_folder, use_case_name)):
                print("NLPGuard: Warning, specified use_case_name already exists in output folder.")
                ts = time.time()
                timestamp = datetime.datetime.fromtimestamp(ts).strftime('%Y%m%d_%H%M%S')
                use_case_name = use_case_name + "_" + timestamp

        use_case_folder_path = os.path.join(output_folder, use_case_name)
        os.mkdir(use_case_folder_path)

        # Create the explainer output folders
        explainer_output_folder = os.path.join(use_case_folder_path, "explainer_outputs")
        os.mkdir(explainer_output_folder)
        os.mkdir(os.path.join(explainer_output_folder, "local_explanations"))
        os.mkdir(os.path.join(explainer_output_folder, "global_explanations"))

        # Create the identifier output folders
        identifier_output_folder = os.path.join(use_case_folder_path, "identifier_outputs")
        os.mkdir(identifier_output_folder)

        # Create the moderator output folders
        moderator_output_folder = os.path.join(use_case_folder_path, "moderator_outputs")
        os.mkdir(moderator_output_folder)

        print(f"NLPGuard: Initialized output folder - {use_case_folder_path}.")
        return output_folder, use_case_name, explainer_output_folder, identifier_output_folder, moderator_output_folder

    def run_full_mitigation_pipeline(self, use_case_name=None, output_folder=None):
        # TODO: implement full mitigation pipeline in one function (explainer + identifier + moderator)
        return

    def run_explainer(self, model, tokenizer, texts, label_ids_to_explain,  explainer_method="integrated-gradients", batch_size=128, device="cpu") -> dict:
        """ Runs the explainer on the texts.

        Args:
            model (:obj:`transformers.AutoModelForSequenceClassification`): The model to explain.
            tokenizer (:obj:`transformers.AutoTokenizer`): The tokenizer to use to preprocess the inputs.
            texts (:obj:`list[str]`): The list of texts to explain.
            label_ids_to_explain (:obj:`list(str)`): The list of label ids to explain.
            explainer_method (:obj:`str`): The explainer method to use. The default is Integrated Gradients.
            device (:obj:`str`, `optional`): The device to run the explainer on. The default is "cpu".

        Returns:
            :obj:`dict`: Dictionary containing the local explanations for each text.
        """

        if explainer_method == "integrated-gradients":
            print("NLPGuard (Explainer): Instantiated Integrated Gradients Explainer")
            explainer = IntegratedGradientsExplainer(model, tokenizer, device)
        elif explainer_method == "shap":
            print("NLPGuard (Explainer): Instantiated SHAP Explainer")
            explainer = ShapExplainer(model, tokenizer, device)
        else:
            print("NLPGuard (Explainer): Unknown Explainer method. Please select ....")

        # Predict labels from texts
        print("NLPGuard (Explainer): Running Predictions")
        df_predictions = self.run_predictions(model, tokenizer, texts, batch_size, device)

        print("NLPGuard (Explainer): Running Local Explanations")
        local_explanations_folder = os.path.join(self.explainer_output_folder, "local_explanations")
        explainer.local_explanations(df_predictions, local_explanations_folder, label_ids_to_explain, self.id2label, batch_size)

        print("NLPGuard (Explainer): Running Global Explanations")
        output_dict = explainer.global_explanations(label_ids_to_explain, self.id2label, self.explainer_output_folder)

        return output_dict

    def run_identifier(self, output_dict, identifier_method="chatgpt", number_most_important_words=400,
                       hf_token="", hf_endpoint="https://api-inference.huggingface.co/models/meta-llama/Meta-Llama-3-8B-Instruct",
                       device='cuda' if torch.cuda.is_available() else 'cpu'):
        """ Runs the Identifier Component to determine which of the most important words are protected attributes.

        Args:
            output_dict (:obj:dict): Dictionary containing the most important words for each label.
            identifier_method (str, optional): Method to use for identifying protected attributes. Defaults to "chatgpt".
            number_most_important_words (int, optional): Number of most important words to use for identifying protected attributes. Defaults to 400.
        """
        if identifier_method == "chatgpt":
            # Extracting distinct words from most important words for each label
            distinct_words = list(set(word for words_list in output_dict.values() for word in words_list[:number_most_important_words]))

            print("NLPGuard (Identifier): Running ChatGPT Identification of Protected Attributes")
            # Instantiating the ChatGPT implementation of the Identifier
            identifier = ChatGPTIdentifier()

            # Annotating the protected attributes with ChatGPT
            df_annotated, protected_attributes = identifier.annotate_protected_attributes(distinct_words)
        elif identifier_method == "llama":
            # Extracting distinct words from most important words for each label
            distinct_words = list(set(word for words_list in output_dict.values() for word in words_list[:number_most_important_words]))

            print("NLPGuard (Identifier): Running LLaMa Identification of Protected Attributes")
            # Instantiating the ChatGPT implementation of the Identifier
            identifier = LLamaIdentifier(hf_endpoint=hf_endpoint, hf_token=hf_token)

            # Annotating the protected attributes with ChatGPT
            df_annotated, protected_attributes = identifier.annotate_protected_attributes(distinct_words)
        elif identifier_method == "mturk":
            print("NLPGuard (Identifier): Running MTurk Identification of Protected Attributes")
        else:
            print("Unknown Identifier method")

        protected_attributes_dict = {}
        # Printing the protected attributes for each label
        print("NLPGuard (Identifier): Protected Attributes per Label:")
        for output_dict_key in output_dict.keys():
            protected_attributes_current_label = [word for word in output_dict[output_dict_key] if word in protected_attributes]
            print(f"Label {output_dict_key} - Protected Attributes: {protected_attributes_current_label}")
            protected_attributes_dict[self.label2id[output_dict_key]] = protected_attributes_current_label
        return df_annotated, protected_attributes, protected_attributes_dict

    def run_moderator(self, df_train, tokenizer, protected_attributes_per_label_dict, text_column_name, label_column_name, mitigation_strategy="word_removal", mitigate_each_label_separately=False, batch_size=128, n_synonyms=5,
                                                                                 keep_original_sentence=True,):
        """ Runs the Moderator Component to produce a new mitigated training dataset based on the identified protected attributes.
        Args:
            df_train (:obj:`pandas.DataFrame`): the training dataset to mitigate.
            tokenizer (:obj:`transformers.AutoTokenizer`): the tokenizer to use for tokenizing the text.
            protected_attributes_per_label_dict (:obj:dict): a dictionary containing the protected attributes for each label.
            text_column_name (str): the name of the column containing the text.
            label_column_name (str): the name of the column containing the labels.
            mitigation_strategy (str, optional): the mitigation strategy to use. Defaults to "word_removal".
            mitigate_each_label_separately (bool, optional): whether to mitigate each label separately.
            batch_size (int, optional): the batch size to use for the mitigation strategy. Defaults to 128.
        """
        moderator = PandasDataFrameModerator()

        if mitigation_strategy == "word_removal":
            print("NLPGuard (Moderator): Running Word Removal Mitigation Strategy")
            df_train_mitigated = moderator.words_removal_mitigation_strategy(df_train, tokenizer, protected_attributes_per_label_dict, text_column_name, label_column_name, self.id2label, mitigate_each_label_separately, batch_size)
        elif mitigation_strategy == "sentence_removal":
            print("NLPGuard (Moderator): Running Sentence Removal Mitigation Strategy")
            df_train_mitigated = moderator.sentences_removal_mitigation_strategy(df_train,
                                                                                 tokenizer,
                                                                                 protected_attributes_per_label_dict,
                                                                                 text_column_name,
                                                                                 label_column_name,
                                                                                 self.id2label,
                                                                                 mitigate_each_label_separately, batch_size)

        elif mitigation_strategy == "word_replacement_with_synonym":
            print("NLPGuard (Moderator): Running Word Replacement with Synonyms Mitigation Strategy")
            df_train_mitigated = moderator.word_replacement_with_synonyms_mitigation_strategy(df_train,
                                                                                 tokenizer,
                                                                                 protected_attributes_per_label_dict,
                                                                                 text_column_name,
                                                                                 label_column_name,
                                                                                 self.id2label,
                                                                                 n_synonyms,
                                                                                 keep_original_sentence,
                                                                                 mitigate_each_label_separately, batch_size)
        elif mitigation_strategy == "word_replacement_with_hypernym":
            print("NLPGuard (Moderator): Running Word Replacement with Hypernyms Mitigation Strategy (TODO)")
        else:
            print("NLPGuard (Moderator): Unknown Mitigation Strategy. Please select ....")
        return df_train_mitigated

    def run_predictions(self, model, tokenizer, texts, batch_size, device="cpu"):
        """ Runs predictions on the corpus of texts on which compute the explanations.
        Args:
            model (:obj:`transformers.AutoModelForSequenceClassification`): the model to use for predictions.
            tokenizer (:obj:`transformers.AutoTokenizer`): the tokenizer to use for tokenizing the text.
            texts (:obj:List[str]): the list of texts to predict.
            batch_size (int): the batch size to use for predictions.
            device (str, optional): the device to use for predictions. Defaults to "cpu".
        """
        tokenizer_kwargs = {'padding': True, 'truncation': True, 'max_length': 128}
        pipe = pipeline("text-classification", model=model, tokenizer=tokenizer, device=device)

        # TODO - Improve this code using yeld
        preds = []
        iterations = len(texts) // batch_size
        remainder = len(texts) % batch_size
        for i in tqdm(range(iterations)):
            batch_preds = pipe(texts[i:i + batch_size], **tokenizer_kwargs)
            preds.extend(batch_preds)

        batch_preds = pipe(texts[-remainder:], **tokenizer_kwargs)
        preds.extend(batch_preds)

        pred_label_names = [d["label"] for d in preds]
        pred_scores = [d["score"] for d in preds]

        pred_dict = {"text": texts, "pred_label_name": pred_label_names,
                     "pred_label_id": [self.label2id[l] for l in pred_label_names], "pred_score": pred_scores,
                     "pred_probabilities": preds}

        df = pd.DataFrame(pred_dict)
        df.to_csv(os.path.join(self.output_folder, self.use_case_name, "predictions.csv"))
        return df


from abc import ABC, abstractmethod
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM, AutoModelForCausalLM
from tqdm import tqdm
import pandas as pd
import time
import datetime
import openai
import torch
import requests


class Identifier(ABC):
    """ Abstract Moderator Class. """
    def __init__(self):
        return

    @abstractmethod
    def annotate_protected_attributes(self, **kwargs):
        """ Abstract method that annotates the protected attributes. """

        return


class ChatGPTIdentifier(Identifier):
    """ ChatGPT Identifier Class. This class is responsible for identifying protected attributes using ChatGPT.

    Attributes:
        openai_api_key :obj:`str`: OpenAI API Key.
    """

    def __init__(self, openai_api_key=""):
        super().__init__()
        openai.api_key = openai_api_key
        return

    def annotate_protected_attributes(self, tokens, temperature=0.3):
        """  Annotates the protected attributes using ChatGPT.

        Args:
            tokens (:obj:`list(str)`): List of tokens to be annotated.
            temperature (:obj:`float`, optional): Sampling temperature. Defaults to 0.3.

        Returns:
            :obj:`tuple(list(dict), list(str))`: Tuple containing the annotated results and exception logs.
        """

        result = []
        raw_responses = []
        exception_logs = []

        for tk in tqdm(tokens):
            chatgpt_response = ""
            try:
                print(tk)
                chatgpt_response = self._chatgpt_annotate(tk)
                print(chatgpt_response)
                raw_responses.append(f"{tk}: {chatgpt_response}\n")
                chatgpt_main_category, chatgpt_category_main_score, chatgpt_explanation = chatgpt_response.split('|')

                result.append({'word': tk,
                               'chatgpt_protected_category': chatgpt_main_category,  # Protected category annotated by chatgpt
                               'chatgpt_protected_category_score': chatgpt_category_main_score,  # Reliability score of the protected category
                               'chatgpt_explanation': chatgpt_explanation  # Explanation provided for tha annotated category
                               })
            except Exception as ex:
                try:
                    chatgpt_response = chatgpt_response.split('\n')[2]
                    chatgpt_main_category, chatgpt_category_main_score, chatgpt_explanation = chatgpt_response.split('|')
                    result.append({'word': tk,
                                   'chatgpt_protected_category': chatgpt_main_category,  # Protected category annotated by chatgpt
                                   'chatgpt_protected_category_score': chatgpt_category_main_score,  # Reliability score of the protected category
                                   'chatgpt_explanation': chatgpt_explanation  # Explanation provided for tha annotated category
                                   })
                except Exception as ex:
                    print(f"{ex} in {tk}!")
                    print({chatgpt_response})
                    print("\n")
                    exception_logs.append(f"Exception in: {tk}: {chatgpt_response}\n")

        ts = time.time()
        timestamp = datetime.datetime.fromtimestamp(ts).strftime('%Y%m%d_%H%M%S')
        filename = f"ChatGPT-annotation-{timestamp}"

        df_annotated = pd.DataFrame(result)

        df_cleaned = self._clean_chatgpt_responses(df_annotated)

        result, acc_list = self._aggregate_protected_categories_votes(df_cleaned)

        print(result)

        protected_attributes = result.loc[result.count_pa > result.count_non_pa].word.tolist()



        print(f"Identifier - Identified Protected Attributes: {protected_attributes}")

        # TODO - Add logic to save annotation and create a dynamic dictionary of protected attributes
        #df_annotated.to_csv(f"chatgpt_annotation/{filename}.csv", index=None)

        #with open(f'chatgpt_annotation/raw_responses/{filename}.txt', 'w') as responses_file:
        #    responses_file.writelines(raw_responses)

        #with open(f'chatgpt_annotation/log/{filename}.txt', 'w') as responses_file:
        #    responses_file.writelines(exception_logs)

        return result, protected_attributes

    def _clean_chatgpt_responses(self, df_raw,  protected_category_column_name="chatgpt_protected_category"):
        """ Cleans the chatgpt response.

        Args:
            df_raw (:obj:`pd.DataFrame`): Raw dataframe containing the chatgpt response.
            protected_category_column_name (:obj:`str`, optional): Column name containing the protected category. Defaults to "chatgpt_protected_category".

        Returns:
            :obj:`pd.DataFrame`: Cleaned dataframe.
        """

        df = df_raw.copy()
        df["word"] = df["word"].str.lower()
        df["word"] = df["word"].str.strip()
        df[protected_category_column_name] = df[protected_category_column_name].str.lower()
        df[protected_category_column_name] = df[protected_category_column_name].str.strip()

        # Replace the "None" category with "none-category"
        df[protected_category_column_name].replace('none', 'none-category', inplace=True)
        df[protected_category_column_name] = df[protected_category_column_name].fillna("none-category")

        # Correct errors in the response provided by chatgpt for some categories
        df[protected_category_column_name].replace('"pregnancy and maternity', 'pregnancy and maternity', inplace=True)
        df[protected_category_column_name].replace('"disability', 'disability', inplace=True)
        df[protected_category_column_name].replace('"race', 'race', inplace=True)
        return df

    @staticmethod
    def _aggregate_protected_categories_votes(df, protected_category_column_name="chatgpt_protected_category"):
        # Compute the counts in a single grouped operation
        grouped = df.groupby('word')[protected_category_column_name].value_counts().unstack(fill_value=0)

        # Add a total count column
        grouped['count_tot'] = grouped.sum(axis=1)

        # Check if 'none-category' column exists and rename it to 'count_non_pa'
        if 'none-category' in grouped.columns:
            grouped = grouped.rename(columns={'none-category': 'count_non_pa'})
            # Calculate 'count_pa' (which is count_tot - count_non_pa)
            grouped['count_pa'] = grouped['count_tot'] - grouped['count_non_pa']
        else:
            # If 'none-category' column doesn't exist, set 'count_non_pa' to 0 and 'count_pa' to 'count_tot'
            grouped['count_non_pa'] = 0
            grouped['count_pa'] = grouped['count_tot']

        # Reset the index to turn 'word' back into a column
        grouped = grouped.reset_index()

        # Select only the required columns
        result = grouped[['word', 'count_pa', 'count_non_pa', 'count_tot']]

        # Convert to a list of dictionaries, if needed
        acc_list = result.to_dict('records')
        return result, acc_list

    @staticmethod
    def _chatgpt_annotate(tk, temperature=0.3, chatgpt_model="gpt-3.5-turbo") -> dict:
        """ Performs annotation using ChatGPT of the input token.

        Args:
            tk (:obj:`str`): The token to be annotated.
            temperature (:obj:`float`, optional): The temperature of the ChatGPT model. Defaults to 0.3.
            chatgpt_model (:obj:`str`, optional): The ChatGPT model to be used. Defaults to "gpt-3.5-turbo".

        Returns:
            :obj:`dict`: The annotated token.
        """

        completion = openai.chat.completions.create(
            model=chatgpt_model,
            messages=[
                {'role': 'user', 'content': '''Consider these 9 protected categories defined by the Equality Act law to avoid discrimination of automatic decision-making algorithms:
                "Age": A person belonging to a particular age (for example, 32 year olds) or range of ages (for example 18 to 30 year olds).
                "Disability": A person has a disability if she or he has a physical or mental impairment which has a substantial and long-term adverse effect on that person's ability to carry out normal day-to-day activities.
                "Gender reassignment": The process of transitioning from one sex to another.
                "Marriage and civil partnership": Marriage is a union between a man and a woman or between a same-sex couple. Same-sex couples can also have their relationships legally recognised as 'civil partnerships'. Civil partners must not be treated less favourably than married couples.
                "Pregnancy and maternity": Pregnancy is the condition of being pregnant or expecting a baby. Maternity refers to the period after the birth, and is linked to maternity leave in the employment context. In the non-work context, protection against maternity discrimination is for 26 weeks after giving birth, and this includes treating a woman unfavourably because she is breastfeeding.
                "Race": Refers to the protected characteristic of race. It refers to a group of people defined by their race, colour, and nationality (including citizenship) ethnic or national origins.
                "Religion and belief": Religion refers to any religion, including a lack of religion. Belief refers to any religious or philosophical belief and includes a lack of belief. Generally, a belief should affect your life choices or the way you live for it to be included in the definition.
                "Sex": A word that explicitly refers to the gender of a person: e.g., man or woman, she or he, mr or mrs, male of female, madame etc.
                "Sexual orientation": Whether a person's sexual attraction is towards their own sex, the opposite sex or to both sexes.
                '''},
                {'role': 'user', 'content': """You can learn more about the discriminations along each protected characteristic on the following URLs:
                "Age" : https://www.equalityhumanrights.com/en/advice-and-guidance/age-discrimination
                "Disability":https://www.equalityhumanrights.com/en/disability-advice-and-guidance
                "Gender reassignment": https://www.equalityhumanrights.com/en/advice-and-guidance/gender-reassignment-discrimination
                "Marriage and civil partnership": https://www.equalityhumanrights.com/en/advice-and-guidance/marriage-and-civil-partnership-discrimination
                "Pregnancy and maternity": https://www.equalityhumanrights.com/en/node/5916
                "Race": https://www.equalityhumanrights.com/en/advice-and-guidance/race-discrimination
                "Religion and belief": https://www.equalityhumanrights.com/en/religion-or-belief-work
                "Sex": https://www.equalityhumanrights.com/en/advice-and-guidance/sex-discrimination
                "Sexual orientation": https://www.equalityhumanrights.com/en/advice-and-guidance/sexual-orientation-discrimination   """},

                {'role': 'user', 'content': f"""Given the previously defined protected categories "Age", "Disability", "Gender reassignment", "Marriage and civil partnership", "Pregnancy and maternity", "Race", "Religion and belief", "Sex", "Sexual orientation". 
                How would you classify the word "{tk}" and which [0,100] reliability scores (only one) would you give to your assessment? You must assign one category. 
                If a word does not fit any categories, you must assign the category "None" with the reliability score and the relative explanation. 
                Provide the answer in the format: "Protected Category|Reliability Score from 0 to 100 for the protected category|Explanation of why the word belong to the protected category.". 
                In case of a word does not fall into any category, provide the answer in the format: "None|Reliability Score from 0 to 100 for the None category|Explanation of why the word does not fall under any of the defined protected categories. 
                Each answer MUST have exactly two | symbols and only one line; otherwise, I cannot process your response. 
                Please annotate a word to a protected category only if it is strictly related. """}
            ],
            temperature=temperature,
            max_tokens=150
        )
        print(completion)
        return completion.choices[0].message.content


class LLamaIdentifier(Identifier):
    """ The LLamaIdentifier class for identifying protected attributes using LLaMA-based models.

    Attributes:
        hf_endpoint (:obj:`str`): The endpoint of the model on Hugging Face Inference API.
        hf_headers (:obj:`dict`): The headers for the Hugging Face API.
    """
    def __init__(self, hf_endpoint, hf_token):
        """
        Initializes the LLamaIdentifier using the Hugging Face Inference API.

        Args:
            hf_endpoint (str): The endpoint of the model on Hugging Face Inference API.
            hf_token (str): The Hugging Face API token for authentication.
        """

        super().__init__()
        self.hf_endpoint = hf_endpoint
        self.hf_headers = {
            "Authorization": f"Bearer {hf_token}",
            "Content-Type": "application/json",
        }

    def annotate_protected_attributes(self, tokens, temperature=0.3):
        """Annotates tokens with protected attributes using LLaMA-based model.

        Args:
            tokens (:obj:`list(str)`): The list of tokens to be annotated.
            temperature (:obj:`float`, optional): The temperature for sampling from the model. Defaults to 0.3.

        Returns:
            :obj:`list(dict)`: The list of annotated tokens with protected attributes.
        """

        results = []
        raw_responses = []
        exception_logs = []

        print(f"NLPGuard (Identifier): Annotating tokens with protected attributes using {self.hf_endpoint.split('/')[-1]} model")
        for tk in tqdm(tokens):
            try:
                llama_response = self._llama_annotate(tk, temperature)
                raw_responses.append(f"{tk}: {llama_response}\n")

                # Parse response
                protected_category, reliability_score, explanation = llama_response.split('|')
                results.append({
                    'word': tk,
                    'llama_protected_category': protected_category.strip(),
                    'llama_protected_category_score': reliability_score.strip(),
                    'llama_explanation': explanation.strip()
                })
            except Exception as ex:
                exception_logs.append(f"Exception for token '{tk}': {str(ex)}\n")

        # Save results to dataframe and log exceptions
        timestamp = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
        df_annotated = pd.DataFrame(results)

        # Clean and aggregate responses
        df_cleaned = self._clean_llama_responses(df_annotated)
        result, acc_list = self._aggregate_protected_categories_votes(df_cleaned)

        # Log the identified protected attributes
        protected_attributes = result[result.count_pa > result.count_non_pa].word.tolist()
        print(f"LLaMAIdentifier - Identified Protected Attributes: {protected_attributes}")

        return result, protected_attributes

    def _llama_annotate(self, tk, temperature=0.3, prompt_template=None) -> str:
        """Generates annotation for a token using the Hugging Face Inference API.

        Args:
            tk (:obj:`str`): The token to be annotated.
            temperature (:obj:`float`, `optional`): The sampling temperature for the model. Defaults to 0.3.
            prompt_template (:obj:`str`, `optional`): The prompt template for the model. Defaults to None.

        Returns:
            :obj:`str`: The annotation generated by the model.
        """

        if prompt_template is None:
            prompt_template = (
                f"""Consider these 9 protected categories defined by the Equality Act law to avoid discrimination of automatic decision-making algorithms:
                "Age": A person belonging to a particular age (for example, 32 year olds) or range of ages (for example 18 to 30 year olds).
                "Disability": A person has a disability if she or he has a physical or mental impairment which has a substantial and long-term adverse effect on that person's ability to carry out normal day-to-day activities.
                "Gender reassignment": The process of transitioning from one sex to another.
                "Marriage and civil partnership": Marriage is a union between a man and a woman or between a same-sex couple. Same-sex couples can also have their relationships legally recognised as 'civil partnerships'. Civil partners must not be treated less favourably than married couples.
                "Pregnancy and maternity": Pregnancy is the condition of being pregnant or expecting a baby. Maternity refers to the period after the birth, and is linked to maternity leave in the employment context. In the non-work context, protection against maternity discrimination is for 26 weeks after giving birth, and this includes treating a woman unfavourably because she is breastfeeding.
                "Race": Refers to the protected characteristic of race. It refers to a group of people defined by their race, colour, and nationality (including citizenship) ethnic or national origins.
                "Religion and belief": Religion refers to any religion, including a lack of religion. Belief refers to any religious or philosophical belief and includes a lack of belief. Generally, a belief should affect your life choices or the way you live for it to be included in the definition.
                "Sex": A word that explicitly refers to the gender of a person: e.g., man or woman, she or he, mr or mrs, male or female, madame etc.
                "Sexual orientation": Whether a person's sexual attraction is towards their own sex, the opposite sex or to both sexes.
                .\n"""
                f"Classify the word '{tk}' into one of these nine categories or None if it does not belong to any, provide a Reliability Score 0-100, and an explanation for your annotation. "
                f"Provide in format: Protected Category|Reliability Score 0-100|Explanation."
            )

        payload = {
            "inputs": prompt_template,
            "parameters": {"temperature": temperature, "max_length": 150},
        }
        response = requests.post(self.hf_endpoint, headers=self.hf_headers, json=payload)

        if response.status_code == 200:
            return response.json()["generated_text"]
        else:
            raise Exception(f"Failed to query Hugging Face API: {response.text}")

    def _clean_llama_responses(self, df_raw, protected_category_column_name="llama_protected_category") -> pd.DataFrame:
        """ Cleans and standardizes the responses for analysis.

        Args:
            df_raw (:obj:`pd.DataFrame`): The raw dataframe with LLaMA annotations.
            protected_category_column_name (:obj:`str`, `optional`): The name of the protected category column. Defaults to "llama_protected_category".

        Returns:
            :obj:`pd.DataFrame`: The cleaned dataframe.
        """

        df = df_raw.copy()
        df["word"] = df["word"].str.lower().str.strip()
        df[protected_category_column_name] = df[protected_category_column_name].str.lower().str.strip()

        # Replace "None" responses consistently
        df[protected_category_column_name].replace('none', 'none-category', inplace=True)
        df[protected_category_column_name].fillna("none-category", inplace=True)

        # Correct common errors if any
        df[protected_category_column_name].replace('"disability', 'disability', inplace=True)
        df[protected_category_column_name].replace('"race', 'race', inplace=True)

        return df

    @staticmethod
    def _aggregate_protected_categories_votes(df, protected_category_column_name="llama_protected_category") -> pd.DataFrame:
        """ Aggregates the votes (annotations) for each word's assigned categories.

        Args:
            df (:obj:`pd.DataFrame`): The cleaned dataframe with LLaMA annotations.
            protected_category_column_name (:obj:`str`, `optional`): The name of the protected category column. Defaults to "llama_protected_category".

        Returns:
            :obj:`pd.DataFrame`: The aggregated dataframe.
        """
        # Aggregate counts for each word’s assigned categories
        grouped = df.groupby('word')[protected_category_column_name].value_counts().unstack(fill_value=0)
        grouped['count_tot'] = grouped.sum(axis=1)

        # Rename "none-category" to "count_non_pa" and compute "count_pa"
        if 'none-category' in grouped.columns:
            grouped.rename(columns={'none-category': 'count_non_pa'}, inplace=True)
            grouped['count_pa'] = grouped['count_tot'] - grouped['count_non_pa']
        else:
            grouped['count_non_pa'] = 0
            grouped['count_pa'] = grouped['count_tot']

        grouped.reset_index(inplace=True)
        result = grouped[['word', 'count_pa', 'count_non_pa', 'count_tot']]
        acc_list = result.to_dict('records')
        return result, acc_list


from setuptools import setup, find_packages
from pathlib import Path

here = Path(__file__).parent

# Read the contents of your README file
with open(here / 'README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="nlpguard",
    version="0.0.7",
    description="NLPGuard: a Framework for Mitigating the use of Protected Attributes by NLP Classifiers",
    long_description=long_description,  # Use the variable here
    long_description_content_type="text/markdown",
    packages=find_packages(),  # Let find_packages() search the entire project directory
    include_package_data=True,  # This ensures MANIFEST.in is respected
    author="Salvatore Greco",
    author_email="grecosalvatore94@gmail.com",
    url="https://github.com/grecosalvatore/nlpguard",
    install_requires=[
        "numpy>=1.22.4",
        "ferret-xai~=0.4.2",
        "scikit-learn>=0.24.2",
        "matplotlib~=3.7.4",
        "scipy>=1.10.0",
        "tqdm>=4.64.1",
        "setuptools>=58.0.4",
        "openai~=1.3.9",
        "pandas~=2.1.4",
        "torch~=2.1.2",
        "transformers~=4.48.1",
        "nltk>=3.9.1",
        "gensim>=4.3.3",
        "requests~=2.32.3",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
)

setup_requires=['setuptools>=42', 'wheel']
